"""Threshold adaptation strategies."""

from typing import Optional

import numpy as np
from scipy.stats.mstats import mquantiles
from torch import Tensor


class Threshold:

    static = True

    def __init__(self, best_f: float):
        self.best_f = best_f

    def __call__(self, y: Tensor | np.ndarray) -> float:
        return self.best_f


class MaxIncumbent(Threshold):

    static = False

    def __init__(self, eps: Optional[float] = None):
        self.eps = eps if eps is not None else 0.0
        self.prev_best_f = -np.inf

    def __call__(self, y: Tensor | np.ndarray) -> float:
        best_f = max(y.max() + self.eps, self.prev_best_f)
        self.prev_best_f = best_f
        return best_f


class QuantileThreshold(Threshold):

    static = False

    def __init__(self, percentile: float):
        self.percentile = percentile
        self.prev_best_f = -np.inf

    def __call__(self, y: Tensor | np.ndarray) -> float:
        if isinstance(y, Tensor):
            y = y.detach().numpy()
        best_f = float(mquantiles(y, prob=self.percentile).squeeze())
        best_f = max(best_f, self.prev_best_f)
        self.prev_best_f = best_f
        return best_f


class AnnealedThreshold(QuantileThreshold):

    def __init__(self, percentile: float, eta: float):
        super().__init__(percentile=percentile)
        if eta < 0 or eta > 1:
            raise ValueError("eta must be between 0 and 1!")
        self.eta = eta
        self.first_call = True

    def __call__(self, y: Tensor | np.ndarray) -> float:
        if not self.first_call:
            self.percentile = self.percentile**self.eta
        self.first_call = False
        return super().__call__(y)
