"""Utilities for variational search distributions."""

from itertools import batched
from typing import Iterable, NewType, Optional, Sequence

import numpy as np
import torch
from torch import Tensor

SequenceArray = NewType("SequenceArray", Sequence[str])
SequenceTensor = NewType("SequenceTensor", torch.IntTensor)


def batch_indices(
    n: int, batchsize: int, seed: Optional[int] = None
) -> Iterable[np.ndarray]:
    rnd = np.random.RandomState(seed)
    while True:
        rinds = rnd.permutation(n)
        for b in batched(rinds, batchsize):
            yield np.array(b)


#
# Various Sequential NN components
#


class Transpose(torch.nn.Module):

    def __init__(self, dim0=-1, dim1=-2) -> None:
        super().__init__()
        self.dim0 = dim0
        self.dim1 = dim1

    def forward(self, X: Tensor) -> Tensor:
        return X.transpose(dim0=self.dim0, dim1=self.dim1)


class Max(torch.nn.Module):

    def __init__(self, dim: int = -1) -> None:
        super().__init__()
        self.dim = dim

    def forward(self, X: Tensor) -> Tensor:
        return X.max(self.dim)[0]


class Average(torch.nn.Module):

    def __init__(self, dim: int = -1) -> None:
        super().__init__()
        self.dim = dim

    def forward(self, X: Tensor) -> Tensor:
        return X.mean(self.dim)


class RemovePadding(torch.nn.Module):
    """Remove padding on the last dimension of X."""

    def __init__(self, padding_size: int) -> None:
        super().__init__()
        self.psize = padding_size

    def forward(self, X: Tensor) -> Tensor:
        return X[..., self.psize : -self.psize]
